/***************************************************************************
 *   Copyright (C) 2008 by Istvan Bartha   *
 *   bico@k.ro   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "gimper.h"
#include "unhidetrigger.h"

#include <ksystemtray.h>
#include <klocale.h>
#include <kiconloader.h>
#include <kpopupmenu.h>
#include <kapplication.h>
#include <khelpmenu.h>
#include <kaction.h>
#include <kconfig.h>

#include <qtimer.h>
#include <qcursor.h>
#include <qtooltip.h>

#include <kwin.h>
#include <netwm.h>


int border;		// visible border of the hidden Toolbox (in pixels)
int hdelay;		// wait before hiding
int sdelay;		// wait before showing
int hspeed;		// hiding speed (lower is faster)
int sspeed;		// showing speed (lower is faster)

int state;		// -1-startup 0-normal 1-show 2-hide
int xc,yc,w;		// x,y coordinate and width of the Toolbox
int vert;		// vertical desktop size (height)
int horiz;		// horizontal desktop size (width)
int insist;		// activation delay parameter
bool animate;
bool right;
bool starthide;
QValueList<long> tid;	// contains the window ID of the "Toolbox"
QRect trect;		// rectangle for the size of the Toolbox frame
int gimpdesk;		// desktop on which the toolbox is on


UnhideTrigger* UnhideTrigger::the()
{
        static UnhideTrigger UnhideTrigger;
        return &UnhideTrigger;
}

UnhideTrigger::UnhideTrigger()
{
        _timer = new QTimer( this, "UnhideTrigger" );
	connect( _timer, SIGNAL(timeout()), SLOT(pollMouse()) );
}

void UnhideTrigger::setEnabled( bool enable )
{
        if( enable ) {
                enabledCount++;
        } else {
                enabledCount--;
        }

        if ( enabledCount > 0 && !_timer->isActive() ) {
                _timer->start( 100 );
        } else if( enabledCount <= 0 ) {
                _timer->stop();
        }
}

bool UnhideTrigger::isEnabled() const
{
        return _timer->isActive();
}

void UnhideTrigger::updateInfo()
{
    tid = UnhideTrigger::findWindows("", "", "gimp-toolbox", "", 0, false, false);
//    if (KWin::windowInfo( tid[0], NET::WMName ).name()=="GIMP")
    if (KWin::windowInfo( tid[0], 0, NET::WM2WindowRole ).windowRole()=="gimp-toolbox")
    {
	if (state == -1) // detected a new Toolbox
	{
	    KWin::setState( tid[0], NET::KeepAbove );
	    KWin::setState( tid[0], NET::SkipTaskbar );
	}
	state=0;
    } else { state=-1; xc=0; return; }

    trect = KWin::windowInfo( tid[0], NET::WMFrameExtents ).frameGeometry();
    yc=trect.top(); w=trect.width();
//    printf("Toolbox ID = %li Desk: %d Width: %d Top: %d \n",tid[0],gimpdesk,w,yc); // DEBUG
}

void UnhideTrigger::pollMouse()
{
    QPoint pos = QCursor::pos();
    for(int s = 0; s < QApplication::desktop()->numScreens(); s++)
    {
//	printf("State: %d Toolbox ID = %li X:%d PX:%d\n",state,tid[0],xc,pos.x()); // DEBUG
	
        if (state==-1) UnhideTrigger::updateInfo(); // look for the Toolbox window ID first

if (!right) { // ---------------- LEFT ------------------

        if (state==0 && xc>-w+border && pos.x()>w) // if the user leaves the area of the Toolbox then 'autohide'
	{
	    UnhideTrigger::updateInfo();
	    if (state == -1) return;
	    state=2; xc=trect.left(); starthide=true;
	    if (!animate) { xc=-w+10+border; KWin::clearState( tid[0], NET::KeepAbove ); }

	    UnhideTrigger::autohide();
	}

        if (state==2 && pos.x() < w) // if the user changes his mind and wants to STOP the 'autohide'
	{
//	    printf("STOP Hide\n"); // DEBUG
	    state=1;
	    if (!animate) xc=-10;
            UnhideTrigger::autoshow();
	}
	
        if (state==0 && xc<0 && pos.x() == 0 && pos.y() > 80 && pos.y() < vert-80) // if the cursor touches the left side then 'autoshow'
	{
	    gimpdesk = KWin::windowInfo( tid[0], NET::WMDesktop ).desktop();
	    if (KWin::currentDesktop() != gimpdesk && gimpdesk != -1) return; // if not on the same desktop with Gimp
	    if (insist == 0)
	    {
		insist = 1;
		QTimer::singleShot( sdelay, this, SLOT(showDelay()) ); // wait before showing
	    }

	    if (insist == 2)
	    {
		insist = 0;
		UnhideTrigger::updateInfo();
		if (state == -1) return;
		state=1; xc=-w+border;
		if (!animate) { xc=-10; KWin::setState( tid[0], NET::KeepAbove ); }

		KWin::forceActiveWindow( tid[0] );
		UnhideTrigger::autoshow();
	    }
	} else { if (insist == 2) insist=0; }

        if (state==1 && pos.x() > w) // if the user changes his mind and wants to STOP the 'autoshow'
	{
//	    printf("STOP Show\n"); // DEBUG
	    state = 0;
	}

} else { // ---------------- RIGHT ------------------

        if (state==0 && xc<horiz-border && pos.x()<horiz-w) // if the user leaves the area of the Toolbox then 'autohide'
	{
	    UnhideTrigger::updateInfo();
	    if (state == -1) return;
	    state=2; xc=trect.left(); starthide=true;
	    if (!animate) { xc=horiz-10-border+1; KWin::clearState( tid[0], NET::KeepAbove ); }

	    UnhideTrigger::autohideRight();
	}

        if (state==2 && pos.x() > horiz-w) // if the user changes his mind and wants to STOP the 'autohide'
	{
//	    printf("STOP Hide\n"); // DEBUG
	    state=1;
	    if (!animate) xc=horiz-w+10+1;
            UnhideTrigger::autoshowRight();
	}
	
        if (state==0 && xc>horiz-w+1 && pos.x() == horiz && pos.y() > 80 && pos.y() < vert-80) // if the cursor touches the left side then 'autoshow'
	{
	    gimpdesk = KWin::windowInfo( tid[0], NET::WMDesktop ).desktop();
	    if (KWin::currentDesktop() != gimpdesk && gimpdesk != -1) return; // if not on the same desktop with Gimp
	    if (insist == 0)
	    {
		insist = 1;
		QTimer::singleShot( sdelay, this, SLOT(showDelay()) ); // wait before showing
	    }

	    if (insist == 2)
	    {
		insist = 0;
		UnhideTrigger::updateInfo();
		if (state == -1) return;
		state=1; xc=horiz-border;
		if (!animate) { xc=horiz-w+10+1; KWin::setState( tid[0], NET::KeepAbove ); }

		KWin::forceActiveWindow( tid[0] );
		UnhideTrigger::autoshowRight();
	    }
	} else { if (insist == 2) insist=0; }

        if (state==1 && pos.x() < horiz-w) // if the user changes his mind and wants to STOP the 'autoshow'
	{
//	    printf("STOP Show\n"); // DEBUG
	    state = 0;
	}
}

    }
}

void UnhideTrigger::showDelay()
{
	insist = 2;
}

void UnhideTrigger::showToolbox()
{
    NETRootInfo info( qt_xdisplay(), 0 );
    info.moveResizeWindowRequest( tid[0], 1 | ( 1 << 8 ) | ( 1 << 9 ) | ( 2 << 12 ), xc, yc, 0, 0 );

    if (state==1)
    {
	if (xc < 0)
	{
	    UnhideTrigger::autoshow();
	} else state=0;
    }
}
void UnhideTrigger::autoshow()
{
//    printf("SX: %d ",xc); // DEBUG
    xc=xc+10; if (xc>0) xc=0; 
    QTimer::singleShot( sspeed, this, SLOT(showToolbox()) );
}

void UnhideTrigger::hideToolbox()
{
    NETRootInfo info( qt_xdisplay(), 0 );
    info.moveResizeWindowRequest( tid[0], 1 | ( 1 << 8 ) | ( 1 << 9 ) | ( 2 << 12 ), xc, yc, 0, 0 );

    if (state==2)
    {
	if (xc > -w+border)
	{
    	    UnhideTrigger::autohide();
	} else state=0;
    }
}
void UnhideTrigger::autohide()
{
//    printf("HX: %d ",xc); // DEBUG
    if (starthide)
    {
	xc=xc-10; starthide=false;
	QTimer::singleShot( hdelay, this, SLOT(hideToolbox()) ); // wait before hiding
    }
    else
    {
	xc=xc-10; if (xc <= -w) xc = -w+border;
	QTimer::singleShot( hspeed, this, SLOT(hideToolbox()) ); // hiding speed
    }
}

void UnhideTrigger::showToolboxRight()
{
    NETRootInfo info( qt_xdisplay(), 0 );
    info.moveResizeWindowRequest( tid[0], 1 | ( 1 << 8 ) | ( 1 << 9 ) | ( 2 << 12 ), xc, yc, 0, 0 );

    if (state==1)
    {
	if (xc > horiz-w+1)
	{
	    UnhideTrigger::autoshowRight();
	} else state=0;
    }
}
void UnhideTrigger::autoshowRight()
{
//    printf("SRX: %d %d %d %d \n",xc, horiz, w, horiz-w); // DEBUG
    xc=xc-10; if (xc<horiz-w+1) xc=horiz-w+1; 
    QTimer::singleShot( sspeed, this, SLOT(showToolboxRight()) );
}

void UnhideTrigger::hideToolboxRight()
{
    NETRootInfo info( qt_xdisplay(), 0 );
    info.moveResizeWindowRequest( tid[0], 1 | ( 1 << 8 ) | ( 1 << 9 ) | ( 2 << 12 ), xc, yc, 0, 0 );

    if (state==2)
    {
	if (xc < horiz-border)
	{
    	    UnhideTrigger::autohideRight();
	} else state=0;
    }
}
void UnhideTrigger::autohideRight()
{
//    printf("HX: %d ",xc); // DEBUG
    if (starthide)
    {
	xc=xc+10; starthide=false;
	QTimer::singleShot( hdelay, this, SLOT(hideToolboxRight()) ); // wait before hiding
    }
    else
    {
	xc=xc+10; if (xc >= horiz) xc = horiz-border+1;
	QTimer::singleShot( hspeed, this, SLOT(hideToolboxRight()) ); // hiding speed
    }
}

// ----------------------------------

void UnhideTrigger::resetTriggerThrottle()
{
//        _lastTrigger = None;
//        _lastXineramaScreen = -1;
}

void UnhideTrigger::emitTrigger( Trigger t, int XineramaScreen )
{
        if( _lastTrigger == t  && _lastXineramaScreen == XineramaScreen) return;
        resetTriggerThrottle();
        emit triggerUnhide( t, XineramaScreen );
}

void Gimper::init() // --- Some initialization ---
{
    QDesktopWidget *d = QApplication::desktop();
    vert = d->height();    // returns desktop height
    horiz = d->width()-1;    // returns desktop width

    state = -1; w = 200;

    connect(UnhideTrigger::the(), SIGNAL(triggerUnhide(UnhideTrigger::Trigger,int)), this, SLOT(unhideTriggered(UnhideTrigger::Trigger)));
    UnhideTrigger::the()->setEnabled( true );
}

const int SUPPORTED_WINDOW_TYPES_MASK = NET::NormalMask;

QValueList< long > UnhideTrigger::findWindows( QString caption, QString wclass, QString role, QString hostname, int pid, bool currentDesktop, bool normal )
{
    NETRootInfo info( qt_xdisplay(), NET::ClientListStacking );
    QValueList< long > ret;
    long mask1 = 0;
    long mask2 = 0;
    if( !caption.isEmpty() ) 	mask1 |= NET::WMName;
    if( !wclass.isEmpty() )	mask2 |= NET::WM2WindowClass;
    if( !role.isEmpty() )	mask2 |= NET::WM2WindowRole;
    if( !hostname.isEmpty() )	mask2 |= NET::WM2ClientMachine;
    if( currentDesktop )	mask1 |= NET::WMDesktop;
    if( normal )		mask1 |= NET::WMWindowType;
    for( int i = info.clientListStackingCount() - 1; i >= 0; --i )
    {
        WId w = info.clientListStacking()[ i ];
        KWin::WindowInfo winfo = KWin::windowInfo( w, mask1, mask2 );
        if( !winfo.valid()) continue;
        if( !caption.isEmpty() && !winfo.name().contains( caption, false )) continue;
        if( !wclass.isEmpty() && !( winfo.windowClassName() + ' ' + winfo.windowClassClass()).contains( wclass, false )) continue;
        if( !role.isEmpty() && !winfo.windowRole().contains( role, false )) continue;
        if( !hostname.isEmpty() && !winfo.clientMachine().contains( hostname, false )) continue;
        if( currentDesktop && !winfo.isOnCurrentDesktop()) continue;
        if( normal )
        {
            NET::WindowType type = winfo.windowType( SUPPORTED_WINDOW_TYPES_MASK );
            if( type != NET::Unknown && type != NET::Normal ) continue;
        }
        if( pid != 0 )
        {
            NETWinInfo winfo2( qt_xdisplay(), w, qt_xrootwin(), NET::WMPid );
            if( winfo2.pid() != pid ) continue;
        }
	ret.append( w );
    }
    return ret;
}


void Gimper::mouseReleaseEvent( QMouseEvent * e)
{
    if (e->state() & QMouseEvent::MidButton)
    {}
    if (e->state() & QMouseEvent::RightButton)
    {
//	QApplication::exit();
    }
}

void Gimper::mouseDoubleClickEvent( QMouseEvent * e)
{
      Gimper::launch();
}

void Gimper::launch()
{
      gimprocess = new QProcess( this );
      gimprocess->addArgument("gimp");
      gimprocess->start();
}

void Gimper::unhideTriggered(UnhideTrigger::Trigger tr)
{
//    printf("Event \n"); // DEBUG
}

Gimper::Gimper(): KSystemTray( 0, "Gimper" )
{
    KConfig *cfg = kapp->config();
    cfg->setGroup("Options");
//    title = cfg->readEntry("title","Toolbox");	// name of the window to hide
    animate = cfg->readBoolEntry("animate",true);
    right = cfg->readBoolEntry("right",false);
    border = cfg->readNumEntry("border",3);	// visible border of the hidden Toolbox (in pixels)
    hdelay = cfg->readNumEntry("hidedelay",500);// wait before hiding
    sdelay = cfg->readNumEntry("showdelay",200);// wait before showing
    hspeed = cfg->readNumEntry("hidespeed",5);	// hiding speed (lower is faster)
    sspeed = cfg->readNumEntry("showspeed",5);	// showing speed (lower is faster)

    setPixmap( BarIcon( "gimp", 22 ) );
    QToolTip::add(this, i18n("Gimper - the GIMP activator"));

    KPopupMenu *p = contextMenu();
//    p->insertTitle( BarIcon( "gimp", 22 ), kapp->caption() );

    p->insertItem(i18n("Launch The GIMP"), this, SLOT(launch()));

    KRadioAction *r1 = new KRadioAction(i18n("Left"), "", 0, this, SLOT(sideSwitched()), actionCollection(), "type0");
    KRadioAction *r2 = new KRadioAction(i18n("Right"), "", 0, this, SLOT(sideSwitched()), actionCollection(), "type1");
    r1->plug(p);
    r2->plug(p);
    if (right) r2->setChecked(true); else r1->setChecked(true);

    const QString conn = QString::fromLatin1("connection_type");
    r1->setExclusiveGroup(conn);
    r2->setExclusiveGroup(conn);

//    KToggleAction *offscreenitem = new KToggleAction(i18n("Move Offscreen"), "", 0, 0, 0, actionCollection(), "kconnections_notify");
//    connect(offscreenitem, SIGNAL(toggled(bool)), this, SLOT(animateToggled(bool)));
//    offscreenitem->plug(p);
//    offscreenitem->setChecked(false);

    KToggleAction *animateitem = new KToggleAction(i18n("Animate Movement"), "", 0, 0, 0, actionCollection(), "notify");
    connect(animateitem, SIGNAL(toggled(bool)), this, SLOT(animateToggled(bool)));
    animateitem->plug(p);
    animateitem->setChecked(animate);
//    animateitem->setEnabled(false);

    p->insertSeparator();

    p->insertItem(i18n("Help"), (new KHelpMenu(this, kapp->aboutData()))->menu());
}

void Gimper::sideSwitched()
{
    if (right) right = false; else right = true;
}

void Gimper::animateToggled(bool a)
{
    animate = a;
    if (animate) KWin::setState( tid[0], NET::KeepAbove );
}

void Gimper::closeEvent( QCloseEvent* ce ) // -- Some finishing up ---
{
    printf("Exiting ...\n");
    if (state!=-1)
    {
	NETRootInfo info( qt_xdisplay(), 0 );
	if (!right) {
	    info.moveResizeWindowRequest( tid[0], 1 | ( 1 << 8 ) | ( 1 << 9 ) | ( 2 << 12 ), 0, yc, 0, 0 );
	} else {
	    info.moveResizeWindowRequest( tid[0], 1 | ( 1 << 8 ) | ( 1 << 9 ) | ( 2 << 12 ), horiz-w+1, yc, 0, 0 );
	}
	KWin::clearState( tid[0], NET::KeepAbove );
	KWin::clearState( tid[0], NET::SkipTaskbar );
    }
    ce->accept(); // close window

    KConfig *cfg = kapp->config(); // write settings
    cfg->setGroup("Options");
//    cfg->writeEntry("title",title);
    cfg->writeEntry("animate",animate);
    cfg->writeEntry("right",right);
    cfg->writeEntry("border",border);
    cfg->writeEntry("hidedelay",hdelay);
    cfg->writeEntry("showdelay",sdelay);
    cfg->writeEntry("hidespeed",hspeed);
    cfg->writeEntry("showspeed",sspeed);
    cfg->sync();
}


Gimper::~Gimper()
{
}


#include "gimper.moc"
