/***************************************************************************
 *   Copyright (C) 2002 John Firebaugh <jfirebaugh@kde.org>    *
 *   Copyright (C) 2008 by Istvan Bartha   *
 *   bico@k.ro   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __UnhideTrigger_h__
#define __UnhideTrigger_h__

// Fix compilation with --enable-final
#ifdef None
#undef None
#endif


#include <qobject.h>

class UnhideTrigger : public QObject
{
	Q_OBJECT
public:
	enum Trigger { None = 0, Top, TopRight, Right, BottomRight, Bottom, BottomLeft, Left, TopLeft };
	static UnhideTrigger* the();

	void setEnabled( bool enable );
	bool isEnabled() const;

	// this is called whenever an item accepts a trigger, thereby preventing further calling of it again
	void triggerAccepted( UnhideTrigger::Trigger t, int XineramaScreen );
	void resetTriggerThrottle();
	void updateInfo();

        void autoshow();
        void autohide();

        void autoshowRight();
        void autohideRight();

signals:
	void triggerUnhide( UnhideTrigger::Trigger t, int XineramaScreen );
private slots:
	void pollMouse();
	void showToolbox();
	void hideToolbox();

	void showDelay();

	void showToolboxRight();
	void hideToolboxRight();
private:
	UnhideTrigger();
	void emitTrigger( Trigger t , int XineramaScreen );
	Trigger _lastTrigger;
	int _lastXineramaScreen;
	QTimer *_timer;
	int enabledCount;
        QValueList< long > findWindows( QString caption, QString wclass, QString role, QString hostname, int pid, bool currentDesktop, bool normal );
};

#endif
